;Simple Timer with DSM0822B LCD and Optical Rotary Encoder input device
;ATTiny2313 @ 1.8432 MHz crystal, Fuses: low=0xCA, high=0xDB (BOD 2.7V)

;PORT B:
;  PB0: DSM0822B (LCD) SDA
;  PB1: DSM0822B (LCD) SCK
;  PB2: N/C
;  PB3: Enable LCD
;  PB4: Enable Optical Encoder
;  PB5: Encoder Input A
;  PB6: Encoder Input B
;  PB7: Encoder Key Input (active low)
;PORT D:
;  PD0: N/C
;  PD1: N/C
;  PD2: N/C
;  PD3: N/C
;  PD4: N/C
;  PD5: N/C
;  PD6: Speaker (through NPN transistor)

;VERSION HISTORY
;  v0.1  2006-09-13  Arne Rossius
;  v0.5  2006-12-21  Arne Rossius
;  v0.8  2007-01-11  Arne Rossius
;  v0.9  2007-01-29  Arne Rossius
;    * debounce key after power on
;    * power off after connecting batteries

.equ VerMaj = 0
.equ VerMin = 9

.include "../include/tn2313def.inc"

.equ	FREQ = 1843200 ;XTAL clock frequency
.equ	BEEP_NUM = 10 ;number of beeps when finished
.equ	BEEP_LEN_ON = 150 ;duration of each beep
.equ	BEEP_LEN_OFF = 200 ;duration of pause between beeps
.equ	BEEP_FREQ = 87 ;frequency of beeping (lower value means higher frequency!)

.equ	LCD_PORT = PORTB
.equ	LCD_DDR = DDRB
.equ	LCD_PIN = PINB
.equ	LCD_DA = 0
.equ	LCD_CK = 1
.equ	LCD_ENABLE = 3

.equ	ENC_PORT = PORTB
.equ	ENC_PIN = PINB
.equ	ENC_ENABLE = 4
.equ	ENC_A = 5
.equ	ENC_B = 6
.equ	ENC_KEY = 7

.equ	SPK_PORT = PORTD
.equ	SPK_OUT = 6

.def	enc_last = R0
.def	enc_delta = R1

.def	temp = R16
.def	temp2 = R17
.def	count = R18
.def	byte = R19
.def	mask = R20
.def	second = R21
.def	minute = R22
.def	value = R23
.def	maxval = R24
.def	lcdpos = R25
.def	debounce = R26
.def	key = R27
.def	running = R28
.def	subsec = R29
;R30/R31 needed for Z pointer

;==================================================================== INTERRUPTS

.cseg
.org 0x000 ;reset
	rjmp	reset
.org OVF0addr ;interrupt: timer 0 overflow
	rjmp	timer0
.org OVF1addr ;interrupt: timer 1 overflow
	rjmp	timer1
.org PCINTaddr ;pin change interrupt (used for waking up from power down mode)
	reti
	
;=================================================== INTERRUPT: TIMER 0 OVERFLOW

timer0: ;timer 0 overflow (~900 Hz)
	;backup registers
	push	temp
	push	temp2
	push	count
	in	temp, SREG
	push	temp

	;check optical encoder
	ldi	temp, 0
	sbic	ENC_PIN, ENC_B
	ldi	temp, 1
        ;convert gray to binary
	sbis	ENC_PIN, ENC_A
	rjmp	PC+3
	ldi	temp2, 3
	eor	temp, temp2
	;difference new - last
	sub	temp, enc_last
	sbrs	temp, 0
	rjmp	PC+5
	add	enc_last, temp ;store new as next last
	;bit 1 = direction (+/-)	
	andi	temp, 2
	dec	temp
	add	enc_delta, temp

	;debounce encoder key
	sbic	ENC_PIN, ENC_KEY
	rjmp	key_off
	cpi	debounce, 1
	brsh	debounce_end ;key was already pressed
	ldi	debounce, 255 ;key was just pressed
	ldi	key, 1
	rjmp	debounce_end

key_off:
	cpi	debounce, 0 ;key is not pressed
	breq	debounce_end ;debouncing delay has passed
	dec	debounce
debounce_end:
	
	;restore registers and return
	pop	temp
	out	SREG, temp
	pop	count
	pop	temp2
	pop	temp
	reti
	
;=================================================== INTERRUPT: TIMER 1 OVERFLOW

timer1: ;timer 1 overflow (1 Hz)
	;backup registers
	push	temp
	push	temp2
	push	count
	in	temp, SREG
	push	temp

	;reload timer
	ldi	temp, HIGH(65536 - FREQ/1024/4)
	ldi	temp2, LOW(65536 - FREQ/1024/4)
	out	TCNT1H, temp
	out	TCNT1L, temp2
	
	;check key (press >3s to abort countdown)
	inc	debounce
	sbic	ENC_PIN, ENC_KEY
	clr	debounce
	cpi	debounce, 3
	brlo	timer1_key_end
	ldi	running, 0
	rjmp	timer1_end
timer1_key_end:

	;rotating animation
	rcall	lcd_init
	ldi	temp, 7*4
	rcall	lcd_send
	ldi	temp, '\' ;step 1
	cpi	subsec, 3
	brne	PC+2
	ldi	temp, '|' ;step 2
	cpi	subsec, 2
	brne	PC+2
	ldi	temp, '/' ;step 3
	cpi	subsec, 1
	brne	PC+2
	ldi	temp, '-' ;step 4
	rcall	lcd_chr
	rcall	lcd_stop

	;decrement countdown
	dec	subsec
	brne	timer1_end
	ldi	subsec, 4
	mov	value, second
	rcall	bcddec
	mov	second, value
	cpi	second, 0x60
	brlo	timer1_disp

	mov	value, minute
	rcall	bcddec
	mov	minute, value
	ldi	second, 0x59

timer1_disp:
	;display new time
	rcall	disptime

	;check if countdown has finished
	cpi	minute, 0x00
	brne	timer1_end
	cpi	second, 0x00
	brne	timer1_end
	ldi	running, 0

timer1_end:
	;restore registers and return
	pop	temp
	out	SREG, temp
	pop	count
	pop	temp2
	pop	temp
	reti
	
;========================================================================== INIT

;LCD strings
strHello: .db  "TMR V",VerMaj|0x30,".",VerMin|0x30
strSetMin: .db " MIN 00 "
strSetSec: .db " SEC 00 "
strReady: .db "00-00 OK"
strRunning: .db " 00-00 -"
strDone: .db "  DONE  "

reset:
	;stack pointer
	ldi	temp, LOW(RAMEND)
	out	SPL, temp
	
	;ports
	ldi	temp, 0x1C
	out	DDRB, temp
	ldi	temp, 0x80
	out	PORTB, temp
	ldi	temp, 0x7F
	out	DDRD, temp
	ldi	temp, 0x00
	out	PORTD, temp
	
	;disable analog comparator
	ldi	temp, 1<<ACD
	out	ACSR, temp

	;LCD setup
	rcall	wait50ms
	sbi	LCD_PORT, LCD_ENABLE
	rcall	wait50ms
	ldi	ZH, HIGH(strHello*2)
	ldi	ZL, LOW(strHello*2)
	rcall	flash2lcd
	rcall	wait1s
	rcall	wait1s
	rcall	lcd_clear

	;enable interrupts
	sei

;================================================================== MAIN PROGRAM

start:
	;enable pin change interrupt for key and power off
	ldi	temp, 1<<ENC_KEY
	out	PCMSK, temp
	ldi	temp, 1<<PCIE
	out	GIMSK, temp
	rcall	lcd_clear ;power off LCD
	cbi	LCD_PORT, LCD_ENABLE
	ldi	temp, 1<<SE | 1<<SM0 ;sleep mode: power-down
	out	MCUCR, temp
	sleep

	;POWERED OFF UNTIL KEY IS PRESSED

	;power on: disable pin change interrupt, enable LCD and start over
	ldi	temp, 0x00
	out	GIMSK, temp
	sbi	LCD_PORT, LCD_ENABLE
	ldi	key, 0
	
	;debounce key
	ldi	temp, 0x20
poweron_debounce:
	sbis	ENC_PIN, ENC_KEY
	ldi	temp, 0x20
	dec	temp2
	brne	poweron_debounce
	dec	temp
	brne	poweron_debounce
	
	;read settings from EEPROM
	ldi	ZL, 0x23 ;second
	rcall	eeread
	mov	second, temp
	cpi	temp, 0x60
	brlo	PC+2
	ldi	second, 0x00
	ldi	ZL, 0x42 ;minute
	rcall	eeread
	mov	minute, temp
	cpi	temp, 0xA0
	brlo	PC+2
	ldi	minute, 0x00
	
	;enable encoder
	sbi	ENC_PORT, ENC_ENABLE
	ldi	temp, 1<<CS01 ;Clk/8
	out	TCCR0, temp
	ldi	temp, 1<<TOIE0
	out	TIMSK, temp
	rcall	wait50ms
	clr	enc_delta

countdown_set:
	;set minute
	ldi	ZL, LOW(strSetMin*2)
	ldi	ZH, HIGH(strSetMin*2)
	rcall	flash2lcd
	ldi	lcdpos, 5*4
	ldi	maxval, 0x99
	mov	value, minute
	rcall	setval
	mov	minute, value
	
	;set second
	ldi	ZL, LOW(strSetSec*2)
	ldi	ZH, HIGH(strSetSec*2)
	rcall	flash2lcd
	ldi	lcdpos, 5*4
	ldi	maxval, 0x59
	mov	value, second
	rcall	setval
	mov	second, value
	
	;check for 00:00
	cpi	minute, 0x00
	brne	countdown_start
	cpi	second, 0x00
	brne	countdown_start
	rjmp	countdown_set
	
countdown_start:
	;save values to EEPROM
	ldi	ZL, 0x23 ;second
	mov	temp, second
	rcall	eewrite
	ldi	ZL, 0x42 ;minute
	mov	temp, minute
	rcall	eewrite
	
	;display ready screen
	ldi	ZL, LOW(strReady*2)
	ldi	ZH, HIGH(strReady*2)
	rcall	flash2lcd
	ldi	lcdpos, 0
	mov	value, minute
	rcall	dispval
	ldi	lcdpos, 3*4
	mov	value, second
	rcall	dispval
	
	;disable decoder
	cbi	ENC_PORT, ENC_ENABLE

	;wait for keypress
start_wait:
	cpi	key, 0
	breq	start_wait

	;display time
	ldi	ZL, LOW(strRunning*2)
	ldi	ZH, HIGH(strRunning*2)
	rcall	flash2lcd
	rcall	disptime

	;disable encoder timer interrupt and start countdown timer
	ldi	temp, 0x00
	out	TCCR0, temp
	ldi	running, 1
	ldi	subsec, 4
	clr	debounce
	ldi	temp, HIGH(65536 - FREQ/1024/4)
	ldi	temp2, LOW(65536 - FREQ/1024/4)
	out	TCNT1H, temp
	out	TCNT1L, temp2
	ldi	temp, 1<<CS12 | 1<<CS10 ;Clk/1024
	out	TCCR1B, temp
	ldi	temp, 1<<TOIE1
	out	TIMSK, temp

loop:
	cpi	running, 0 ;just finished -> beep and power off
	breq	finished
	
	;countdown is running, so just sleep until the next timer interrupt occurs
	ldi	temp, 1<<SE ;sleep mode: idle
	out	MCUCR, temp
	sleep
	rjmp	loop

finished:
	;disable countdown timer and timer interrupt
	ldi	temp, 0x00
	out	TIMSK, temp
	ldi	temp, 0x00
	out	TCCR1B, temp

	;display "done" and beep a few times
	ldi	ZL, LOW(strDone*2)
	ldi	ZH, HIGH(strDone*2)
	rcall	flash2lcd

	ldi	count, BEEP_NUM
beep_outer:
	;beep
	ldi	temp, BEEP_LEN_ON
beep_inner_on:
	sbrc	temp, 0
	sbi	SPK_PORT, SPK_OUT
	sbrs	temp, 0
	cbi	SPK_PORT, SPK_OUT
	ldi	temp2, BEEP_FREQ
beep_delay_on:
	nop
	nop
	dec	temp2
	brne	beep_delay_on
	dec	temp
	brne	beep_inner_on
	;do not beep (pause)
	ldi	temp, BEEP_LEN_OFF
beep_inner_off:
	sbrc	temp, 0
	cbi	SPK_PORT, SPK_OUT
	sbrs	temp, 0
	cbi	SPK_PORT, SPK_OUT
	ldi	temp2, BEEP_FREQ
beep_delay_off:
	nop
	nop
	dec	temp2
	brne	beep_delay_off
	dec	temp
	brne	beep_inner_off
	;one beep/pause complete
	dec	count
	brne	beep_outer
	
	rjmp	start

;============================================================ UNIVERSAL ROUTINES

bcdinc:
	;increment packed BCD number
	subi	value, -7
	brhc	PC+2
	subi	value, 6
	ret

;--------------------

bcddec:
	;decrement packed BCD number
	subi	value, 1
	brhc	PC+2
	subi	value, 6
	ret

;--------------------

wait50ms:
	;wait ~50ms
	ldi	temp, 0x80
	ldi	temp2, 0
wait50ms_loop:
	dec	temp2
	brne	wait50ms_loop
	dec	temp
	brne	wait50ms_loop
	ret

;--------------------
	
wait1s:
	;wait ~1s (20*~50ms)
	ldi	count, 20
wait1s_loop:
	rcall	wait50ms
	dec	count
	brne	wait1s_loop
	ret

;=============================================================== EEPROM ROUTINES

eewrite:
	;write temp to the EEPROM at ZL
	sbic	EECR,EEPE
	rjmp	eewrite
	out	EEAR, ZL
	out	EEDR, temp
	sbi	EECR, EEMPE
	sbi	EECR, EEPE
	ret
	
;--------------------

eeread:
	;read temp from the EEPROM at ZL
	sbic	EECR, EEPE
	rjmp	eeread
	out	EEAR, ZL
	sbi	EECR, EERE
	in	temp, EEDR
	ret

;====================================================== OPTICAL ENCODER ROUTINES

setval:
	;set temp between 0 and maxval (all in packed BCD), display it at lcdpos
	rcall	dispval ;display current value
	
set_encwait:
	;wait for encoder action
	cpi	key, 1
	brsh	set_done
	tst	enc_delta
	breq	set_encwait
	brmi	set_dec

	;increment
	cp	value, maxval
	breq	PC+2
	rcall	bcdinc
	dec	enc_delta
	rjmp	setval
	
set_dec:
	;decrement
	cpi	value, 0x00
	breq	PC+2
	rcall	bcddec
	inc	enc_delta
	rjmp	setval

set_done:
	ldi	key, 0
	ret

;============================================================== I2C LCD ROUTINES
	
.macro lcd_clk1
	cbi	LCD_DDR, LCD_CK
.endmacro
	
.macro lcd_clk0
	sbi	LCD_DDR, LCD_CK
.endmacro
	
.macro lcd_dat1
	cbi	LCD_DDR, LCD_DA
.endmacro
	
.macro lcd_dat0
	sbi	LCD_DDR, LCD_DA
.endmacro

;--------------------

flash2lcd:
	;display 8 byte ASCII text at FLASH[Z] on the LCD
	push	count
	rcall	lcd_init ;init LCD
	ldi	temp, 0x00 ;data pointer = 0
	rcall	lcd_send
	ldi	count, 8
flash2lcd_loop:
	lpm	temp, Z+
	push	ZL
	push	ZH
	rcall	lcd_chr
	pop	ZH
	pop	ZL
	dec	count
	brne	flash2lcd_loop
	rcall	lcd_stop
	pop	count
	ret
	
;--------------------

disptime:
	;display remaining time on the LCD
	ldi	lcdpos, 1*4 ;minute
	mov	value, minute
	rcall	dispval
	rcall	lcd_init ;hyphen
	ldi	temp, 3*4
	rcall	lcd_send
	ldi	temp, '-'
	rcall	lcd_chr
	rcall	lcd_stop
	ldi	lcdpos, 4*4 ;second
	mov	value, second
	rcall	dispval
	ret

;--------------------

dispval:
	;display a packed BCD value on the LCD at a given position
	rcall	lcd_init
	mov	temp, lcdpos
	rcall	lcd_send
	mov	temp, value
	swap	temp
	andi	temp, 0x0F
	ori	temp, 0x30
	rcall	lcd_chr
	mov	temp, value
	andi	temp, 0x0F
	ori	temp, 0x30
	rcall	lcd_chr
	rcall	lcd_stop
	ret

;--------------------
	
lcd_clear:
	;clear LCD
	rcall	lcd_init
	ldi	temp, 0x00 ;data pointer = 0
	rcall	lcd_send
	ldi	count, 18 ;send 18 zero bytes
lcd_clear_loop:
	ldi	temp, 0x00
	rcall	lcd_send
	dec	count
	brne	lcd_clear_loop
	rcall	lcd_stop
	ret
	
;------------------

lcd_init:
	;initialize LCD for data transfer
	rcall	lcd_start
	ldi	temp, 0x70 ;select LCD
	rcall	lcd_send
	ldi	temp, 0xE0 ;subdevice 0
	rcall	lcd_send
	ldi	temp, 0xC8 ;switch display on
	rcall	lcd_send
	ret
	
;--------------------
	
lcd_chr:
	;send a character to the LCD
	ldi	ZH, HIGH(charset*2)
	ldi	ZL, LOW(charset*2)
	lsl	temp ;each char is a word!
	add	ZL, temp
	ldi	temp, 0
	adc	ZH, temp
	lpm	temp, Z+ ;low byte
	rcall	lcd_send
	lpm	temp, Z+ ;high byte
	rcall	lcd_send
	ret
	
;--------------------
	
lcd_off:
	;switch LCD off
	rcall	lcd_start
	ldi	temp, 0x70 ;select LCD
	rcall	lcd_send
	ldi	temp, 0xE0 ;subdevice 0
	rcall	lcd_send
	ldi	temp, 0x40 ;switch display off
	rcall	lcd_send
	rcall	lcd_stop
	ret
	
;--------------------
	
lcd_send:
	;send one byte via I2C
	ldi	temp2, 0x80
	mov	mask, temp2
lcd_send_loop:
	lcd_dat1
	mov	temp2, temp
	and	temp2, mask
	brne	lcd_send_clk
	lcd_dat0
lcd_send_clk:
	lcd_clk1
	lcd_clk0
	lsr	mask
	brne	lcd_send_loop
	lcd_dat1
	lcd_clk1
	clt
	sbis	LCD_PIN, LCD_DA
	set
	lcd_clk0
	ret
	
;--------------------
	
lcd_start:
	;I2C start condition
	lcd_dat1
	lcd_clk1
	lcd_dat0
	lcd_clk0
	ret
	
lcd_stop:
	;I2C stop condition
	lcd_clk0
	lcd_dat0
	lcd_clk1
	lcd_dat1
	ret
	
;--------------------

charset: ;7-Bit ASCII charset for 16-Segment LCD
;     0      1      2      3      4      5      6      7      8      9      A      B      C      D      E      F
.dw $0000, $0000, $0000, $0000, $0000, $0000, $0000, $0000, $0000, $0000, $0000, $0000, $0000, $0000, $0000, $0000 ;00
.dw $FFFF, $FFFF, $FFFF, $FFFF, $FFFF, $FFFF, $FFFF, $FFFF, $FFFF, $FFFF, $FFFF, $FFFF, $FFFF, $FFFF, $FFFF, $FFFF ;10
.dw $0000, $D04D, $0104, $A4BC, $BC3D, $4812, $26BD, $0004, $B004, $200D, $6666, $2424, $0002, $0420, $0008, $4002 ;20
.dw $D99B, $4900, $95A9, $9D29, $0D30, $9239, $9CB9, $1901, $9DB9, $9D39, $0009, $0003, $4200, $0420, $0042, $3501 ;30
.dw $BD91, $1DB1, $BD0D, $9099, $B90D, $90B9, $10B1, $9C99, $0DB0, $2004, $8908, $42B0, $8098, $49D0, $0BD0, $9999 ;40
.dw $15B1, $9B99, $17B1, $9C39, $3005, $8998, $4092, $0B92, $4242, $6040, $D00B, $B004, $0240, $200D, $1001, $8008 ;50
.dw $0040, $1DB1, $BD0D, $9099, $B90D, $90B9, $10B1, $9C99, $0DB0, $2004, $8908, $42B0, $8098, $49D0, $0BD0, $9999 ;60
.dw $15B1, $9B99, $17B1, $9C39, $3005, $8998, $4092, $0B92, $4242, $6040, $D00B, $B024, $2004, $240D, $1001, $0000 ;70
